@@ -1,12 +1,12 @@
 package SVN::Notify::HTML::ColorDiff;
 
-# $Id: ColorDiff.pm 3385 2008-02-06 06:05:30Z theory $
+# $Id: ColorDiff.pm 3471 2008-02-26 01:21:28Z theory $
 
 use strict;
 use HTML::Entities;
 use SVN::Notify::HTML ();
 
-$SVN::Notify::HTML::ColorDiff::VERSION = '2.67';
+$SVN::Notify::HTML::ColorDiff::VERSION = '2.70';
 @SVN::Notify::HTML::ColorDiff::ISA = qw(SVN::Notify::HTML);
 
 =head1 Name
@@ -54,34 +54,14 @@ specify C<--handler HTML::ColorDiff>.
 
   $notifier->output_css($file_handle);
 
-This method starts outputs the CSS for the HTML message. It overrides the
-same method on SVN::Notify::HTML to add CSS for the colorized diff.
+This method starts outputs the CSS for the HTML message.
+SVN::Notify::HTML::ColorDiff adds extra CSS to its output so that it can
+nicely style the diff.
 
 =cut
 
-sub output_css {
-    my ($self, $out) = @_;
-    $self->SUPER::output_css($out);
-    print $out
-      qq(#patch h4 {font-family: verdana,arial,helvetica,sans-serif;),
-          qq(font-size:10pt;padding:8px;background:#369;color:#fff;),
-          qq(margin:0;}\n),
-      qq(#patch .propset h4, #patch .binary h4 {margin:0;}\n),
-      qq(#patch pre {padding:0;line-height:1.2em;margin:0;}\n),
-      qq(#patch .diff {width:100%;background:#eee;padding: 0 0 10px 0;),
-          qq(overflow:auto;}\n),
-      qq(#patch .propset .diff, #patch .binary .diff  {padding:10px 0;}\n),
-      qq(#patch span {display:block;padding:0 10px;}\n),
-      qq(#patch .modfile, #patch .addfile, #patch .delfile, #patch .propset, ),
-          qq(#patch .binary, #patch .copfile {border:1px solid #ccc;),
-          qq(margin:10px 0;}\n),
-      qq(#patch ins {background:#dfd;text-decoration:none;display:block;),
-          qq(padding:0 10px;}\n),
-      qq(#patch del {background:#fdd;text-decoration:none;display:block;),
-          qq(padding:0 10px;}\n),
-      qq(#patch .lines, .info {color:#888;background:#fff;}\n);
-    return $self;
-}
+# We use _css() so that ColorDiff can override it and the filters then applied
+# only one to all of the CSS.
 
 ##############################################################################
 
@@ -94,6 +74,11 @@ C<$out_file_handle> for inclusion in the notification message. The diff is
 output with nice colorized HTML markup. Each line of the diff file is escaped
 by C<HTML::Entities::encode_entities()>.
 
+If there are any C<diff> filters, this method will do no HTML formatting, but
+redispatch to L<SVN::Notify::output_diff|SVN::Notify/"output_diff">. See
+L<Writing Output Filters|SVN::Notify/"Writing Output Filters"> for details on
+filters.
+
 =cut
 
 my %types = (
@@ -105,6 +90,9 @@ my %types = (
 
 sub output_diff {
     my ($self, $out, $diff) = @_;
+    if ( $self->filters_for('diff') ) {
+        return $self->SUPER::output_diff($out, $diff);
+    }
     $self->_dbpnt( "Outputting colorized HTML diff") if $self->verbose > 1;
 
     my $in_div;
@@ -182,8 +170,11 @@ sub output_diff {
                 $in_span = '';
             } elsif ($line =~ /^\@\@/) {
                 print $out "</$in_span>" if $in_span;
-                print $out qq{<span class="lines">}, encode_entities($line, '<>&"'),
-                  "\n</span>";
+                print $out (
+                    qq{<span class="lines">},
+                    encode_entities($line, '<>&"'),
+                    "\n</span>",
+                );
                 $in_span = '';
             } elsif ($line =~ /^([-+])/) {
                 my $type = $1 eq '+' ? 'ins' : 'del';
@@ -191,7 +182,11 @@ sub output_diff {
                     print $out encode_entities($line, '<>&"'), "\n";
                 } else {
                     print $out "</$in_span>" if $in_span;
-                    print $out qq{<$type>}, encode_entities($line, '<>&"'), "\n";
+                    print $out (
+                        qq{<$type>},
+                        encode_entities($line, '<>&"'),
+                        "\n",
+                    );
                     $in_span = $type;
                 }
             } else {
@@ -199,8 +194,11 @@ sub output_diff {
                     print $out encode_entities($line, '<>&"'), "\n";
                 } else {
                     print $out "</$in_span>" if $in_span;
-                    print $out qq{<span class="cx">},
-                        encode_entities($line, '<>&"'), "\n";
+                    print $out (
+                        qq{<span class="cx">},
+                        encode_entities($line, '<>&"'),
+                        "\n",
+                    );
                     $in_span = 'span';
                 }
             }
@@ -214,6 +212,31 @@ sub output_diff {
     return $self;
 }
 
+##############################################################################
+
+sub _css {
+    my $css = shift->SUPER::_css;
+    push @$css,
+        qq(#patch h4 {font-family: verdana,arial,helvetica,sans-serif;),
+            qq(font-size:10pt;padding:8px;background:#369;color:#fff;),
+            qq(margin:0;}\n),
+        qq(#patch .propset h4, #patch .binary h4 {margin:0;}\n),
+         qq(#patch pre {padding:0;line-height:1.2em;margin:0;}\n),
+        qq(#patch .diff {width:100%;background:#eee;padding: 0 0 10px 0;),
+            qq(overflow:auto;}\n),
+        qq(#patch .propset .diff, #patch .binary .diff  {padding:10px 0;}\n),
+        qq(#patch span {display:block;padding:0 10px;}\n),
+        qq(#patch .modfile, #patch .addfile, #patch .delfile, #patch .propset, ),
+            qq(#patch .binary, #patch .copfile {border:1px solid #ccc;),
+            qq(margin:10px 0;}\n),
+        qq(#patch ins {background:#dfd;text-decoration:none;display:block;),
+            qq(padding:0 10px;}\n),
+        qq(#patch del {background:#fdd;text-decoration:none;display:block;),
+            qq(padding:0 10px;}\n),
+        qq(#patch .lines, .info {color:#888;background:#fff;}\n);
+    return $css;
+}
+
 1;
 __END__
 
@@ -250,13 +273,13 @@ message body. Maybe use absolute positioning CSS?
 
 =head1 Author
 
-David Wheeler <david@kineticode.com>
+David E. Wheeler <david@kineticode.com>
 
 =head1 Copyright and License
 
 Copyright (c) 2004-2008 Kineticode, Inc. All Rights Reserved.
 
-This module is free software; you can redistribute it and/or modify it under the
-same terms as Perl itself.
+This module is free software; you can redistribute it and/or modify it under
+the same terms as Perl itself.
 
 =cut
